--	main file for LibBuilder

--	NOTE: Revisions < 3 are actually incompatible 
--	but never passed experimental-beta status.


local BASE, VERSION, MINOR = 'LibBuilder', '1.0', 3
if (...):find(BASE..'_dev') == 1 then BASE = ... end
local MAJOR = BASE..'-'..VERSION


local core = LibStub:NewLibrary(MAJOR, MINOR)
if not core then return end


--	external upvalues
local _G = _G
local LibStub = _G.LibStub
local type = _G.type
local select = _G.select


local shellDefaultVersion = '1.0'	--	used if we supply the shell


--	initialize fields
--	leave the client fields intact - clientTable, clientMajor, clientMinor, clientOldMinor

--	Aliases are 'virtual' methods that are only 
--	available through corecall and are not recursive.
core.aliases = {
	[''] = 'LastLib', 	--	shortcut for accessing through shellcall, ie. LibBuilder('1.1', '')
	new = 'NewLib',
	last = 'LastLib',
	close = 'CloseLib',
}

core.errorStrings = {
	TEMPLATE = MAJOR..': %s!', 
	BADVERSION = '%s requires a textual major and numeric minor version', 
	NOLIB = '%s called but no library is being built', 
	LEFTOPEN = '%s, Rev. %s was not closed properly',
	SHELLNODEFAULT = 'The default version is unavailable',
}

--	reuse the old frame if it exists
if core.frame == nil then
	--	Create a frame so we can throw a warning if 
	--	a library is still open when OnUpdate fires.
	--	Use OnUpdate instead of ADDON_LOADED because an error in the client 
	--	library would load Blizzard_DebugTools, confuse the builder, and 
	--	seemed to prevent the real error message from showing.
	local frame = CreateFrame('Frame')
	core.frame = frame
	
	frame:Hide()
	frame:SetScript('OnUpdate', function() core:OnUpdate() end)
end
--	we don't want to hide the frame when upgrading because a client lib may 
--	still be open (which shouldn't happen but this isn't the place to check)


--	Internal; uses frame updates to check if a client library was left open.
--	This should never run if all clients are closed properly.
function core:OnUpdate()
	if not self.clientMajor then
		self.frame:Hide()
		error("This shouldn't happen!")
	end
	
	--	Close the client so the error only happens once.
	--	We have to cache the version info because it'll get wiped.
	local _, major, minor = self:CloseLib()
	
	local message = self:BuildErrorMessage('LEFTOPEN', major, minor)
	error(message)
end


---	Registers a new client library.
--	@name LibBuilder:NewLib
--	@param		major		The major version of the new client library.
--	@param		minor		The minor version of the new client library.
--	@return		libTable	The table for the new library, or nil if no upgrade is necessary.
--	@return		oldMinor	Minor version of the previous instance of the library, if applicable.
function core:NewLib(major, minor)
	--	check for collisions
	if self.clientMajor then
		--	close it properly instead of just 'shortcutting' by overwriting
		local _, oldMajor, oldMinor = self:CloseLib()
		
		--	We can't use error() because we need to return values for the new lib.
		local message = self:BuildErrorMessage('LEFTOPEN', oldMajor, oldMinor)
		geterrorhandler()(message)
	end
	
	--	check the supplied version info
	if type(major) ~= 'string' or not tonumber(minor) then
		local message = self:BuildErrorMessage('BADVERSION', 'NewLib')
		error(message)
	end
	
	--	try LibStub first in case it generates an unexpected error
	local clientTable, oldMinor = LibStub:NewLibrary(major, minor)
	
	self.clientTable, self.clientMajor, self.clientMinor, self.clientOldMinor = clientTable, major, minor, oldMinor
	
	--	show the frame so we can use OnUpdate
	--	to check if this client was left open.
	self.frame:Show()

	return clientTable, major, minor, oldMinor
end


---	Retrieves the table and version info for the current client library.
--	@name LibBuilder:LastLib()
--	@return library		The client library's table, or nil if no upgrade is necessary.
--	@return major		Major version of the client library.
--	@return minor		Minor version of the client library.
--	@return oldMinor	The previous minor version of the client library, if applicable.
function core:LastLib()
	if not self.clientMajor then
		local message = self:BuildErrorMessage('NOLIB', 'LastLib')
		error(message)
	end
	
	return self.clientTable, self.clientMajor, self.clientMinor, self.clientOldMinor
end


---	Closes the active client library. This is used for 
--	tracking, to make sure two libraries don't collide.
--	This function returns the same values as LastLib, 
--	as a shortcut in case the client library needs to 
--	do last-second initialization.
--	@name LibBuilder:CloseLib
--	@see LastLib
function core:CloseLib()
	--	we don't want programmers calling this to suppress 
	--	errors when they register their own library.
	if not self.clientMajor then
		local message = self:BuildErrorMessage('NOLIB', 'CloseLib')
		error(message)
	end
	
	--	grab values before reset so we can return them
	local clientTable, major, minor, oldMinor = self:LastLib()
	
	self.clientTable, self.clientMajor, self.clientMinor, self.clientOldMinor = nil, nil, nil, nil
	
	--	the library was closed properly, so disable the OnUpdate frame
	self.frame:Hide()
	
	return clientTable, major, minor, oldMinor
end


--	Internal; builds an error message.
--	@param	errorName	Name of the error
--	@param	...			List of arguments applicable to the error type.
--	@return message		The complete error string.
function core:BuildErrorMessage(errorName, ...)
	local errorFormat = self.errorStrings[errorName]
	assert(errorFormat)
	
	local body = errorFormat:format(...)
	local message = self.errorStrings.TEMPLATE:format(body)

	return message
end


--	Internal; used for the core's __call metamethod.
--	If the core table is called with no arguments, it 
--	will call the function associated with the '' alias. 
--	If the first argument is a valid key or alias, 
--	the associated value will be called with the remaining 
--	arguments. Otherwise, NewLib will be called 
--	with all arguments.
local function corecall(self, ...)
	if self[...] then												--	if first arg is valid key
		return self[...](self, select(2, ...))						--	call it with any remaining args
	elseif self.aliases[...] or select('#', ...) == 0 then			--	if first is valid alias or no args provided
		return self[self.aliases[... or ''] ](self, select(2, ...)) --	retrieve alias and pass any remaining args
	else															--	if args didn't match anything above
		return self:NewLib(...)										--	assume they were meant to go to NewLib
	end
end
setmetatable(core, { __call = corecall })
_G[MAJOR] = core


--	create the shell if one doesn't already exist
local shell = _G[BASE]		--	we'll reuse this local after the if block
if not shell then
	shell = {}
	_G[BASE] = shell
	
	shell.builtBy = VERSION		--	currently unused
	shell.defaultVersion = shellDefaultVersion
	
	--	shell should act like the default version as much as possible, but we 
	--	make no guarantees that new methods won't be inserted into the shell
	
	--	Internal; if one argument is provided and it is a valid index 
	--	in this table, the matching value will be returned. If 
	--	multiple arguments are provided and the first is a valid 
	--	index, the remaining arugments will be passed as a call 
	--	to the matching value. If the first argument is -not- a 
	--	valid index, all arguments will be passed as a call to 
	--	the default library.
	local function shellcall(self, ...)		--	later revision can replace this if it wants to do anything funky with passing self to methods
		if ... == '' or rawget(self, (...)) then	--	don't use normal indexing, in order to send method calls through libcall
			if select('#', ...) == 1 then
				return self[...]
			else
				return self[...](select(2, ...))
			end
		else
			return self[self.defaultVersion](...)	--	works for no arg too
		end
	end
	
	
	--	Internal; returns the default version if key is 
	--	an empty string, otherwise checks for the key in 
	--	the default library.  Note that lua will not call 
	--	this metamethod if the specified key is present 
	--	in the shell.
	local function shellindex(self, key)
		--	FIXME: apply aliases to shell?
		if key == nil then
			return nil
		end
		
		local defaultLib = rawget(self, self.defaultVersion)
		
		if key == '' then
			return defaultLib		--	may be nil
		else
			if defaultLib == nil then
				--	can't go further if default isn't present
				local message = core:BuildErrorMessage('SHELLNODEFAULT')
				error(message)
			end
			
			--	We'll get around the 'wrong self' issue by 
			--	overloading __newindex to point back to defaultLib.
			--	NOTE: this won't respect aliases, but 
			--	indexing the default library doesn't either.
			return defaultLib[key]
		end
	end
	
	
	local function shellnewindex(self, key, value)
		--	new versions of the library must be added  
		--	with rawset, to bypass this function
		local defaultLib = self[self.defaultVersion]
		
		--	this shouldn't happen, but just in case
		if defaultLib == nil then
			--	can't go further if default isn't present
			local message = core:BuildErrorMessage('SHELLNODEFAULT')
			error(message)
		end
		
		defaultLib[key] = value
	end
	
	
	setmetatable(shell, { __call = shellcall, __index = shellindex, __newindex = shellnewindex})
end

--	add current version to the shell
rawset(shell, VERSION, core)
